import resource

try:
    '''
    Prefer Pytorch's multiprocessing module for optimized sharing of Tensor
    '''
    import torch.multiprocessing as mp
    try:
        resource.setrlimit(resource.RLIMIT_NOFILE, (1048576, 1048576))
    except:
        mp.set_sharing_strategy('file_system')
except:
    import multiprocessing as mp

mp = mp.get_context('spawn')



from functools import partial
from queue import Queue
from threading import Thread
from typing import List, Any, Union

from .pipeline import AsyncPipeline, SyncPipeline
from .resource import Resources
from .stage import Stage
from .task import Task
from .utils import get_logger, progressbar


class Job(object):

    '''
    A job for the system. A job is a wrapper over a task for the first stage.
    '''

    def __init__(self, name: str, task: Task, length: Union[None, int] = None):
        self.name = name
        self.task = task
        self.length = length

    def finish(self, results):
        self.results = results


class System(object):

    '''
    A system consists of multiple peer pipelines.
    '''

    def __init__(self, num_pipeline: int = 1, debug=False,
                 show_progress=True,
                 job_queue_size: int = -1,
                 pipeline_job_queue_size: int = 32,
                 resource_scan_args={}, pipeline_build_args={}):
        self.logger = get_logger(repr(self))
        self.num_pipeline = num_pipeline
        self.pipeline_fn = partial(
            AsyncPipeline, job_queue_size=pipeline_job_queue_size)
        if debug:
            self.logger.info('Debug mode: one SyncPipeline in single process.')
            self.num_pipeline = 1
            self.pipeline_fn = SyncPipeline
        self.show_progress = show_progress
        self.job_queue = Queue(job_queue_size)
        self.resources = Resources(**resource_scan_args)
        self.logger.info('Available resources: %s', repr(self.resources))
        self.result_queue = Queue()
        self.job_count = 0
        self.build(**pipeline_build_args)

    def get_stages(self, resources) -> List[Stage]:
        '''
        Define the stages in a pipeline with given resources.
        '''
        raise NotImplementedError

    def get_results(self, results_gen) -> List[Any]:
        '''
        Define how to extract final results from output tasks.
        '''
        raise NotImplementedError

    def monit_pipeline(self, pipeline_id):
        pipeline = self.pipelines[pipeline_id]
        while True:
            job = self.job_queue.get()
            if job is None:
                return
            pipeline.job_queue.put(job.task)
            pipeline.reset()
            results_gen = pipeline.wait()
            if self.show_progress:
                results_gen = progressbar(
                    results_gen,
                    desc=' Pipeline-%d(%s)' % (pipeline_id, job.name),
                    total=job.length, position=pipeline_id, leave=False)
            results = self.get_results(results_gen)
            job.finish(results)
            self.result_queue.put(job)
            self.progressbar.update()

    def build(self, **pipeline_args):
        self.pipelines = []
        for pipeline_i, resources in enumerate(
                self.resources.split(self.num_pipeline)):
            stages = self.get_stages(resources)
            pipeline = self.pipeline_fn(stages, **pipeline_args)
            self.pipelines.append(pipeline)
            self.logger.info('Building pipeline %d: \n\t%s', pipeline_i,
                             '\n\t'.join([repr(s) for s in stages]))
        self.monitor_threads = [Thread(target=self.monit_pipeline, args=(i,))
                                for i in range(self.num_pipeline)]
        self.progressbar = None

    def start(self):
        self.logger.info('Starting system')
        self.progressbar = progressbar(
            total=self.job_count, desc='Jobs', position=self.num_pipeline)
        for pipeline in self.pipelines:
            pipeline.start()
        for thread in self.monitor_threads:
            thread.start()

    def add_job(self, job):
        self.job_queue.put(job)
        self.job_count += 1
        self.progressbar.total = self.job_count
        self.progressbar.refresh()

    def end(self):
        for pipeline in self.pipelines:
            pipeline.end()
        for pipeline in self.pipelines:
            pipeline.join()
        for _ in range(self.num_pipeline):
            self.job_queue.put(None)
        for thread in self.monitor_threads:
            thread.join()
        self.progressbar.close()

    def terminate(self):
        for pipeline in self.pipelines:
            pipeline.terminate()
        for thread in self.monitor_threads:
            thread.terminate()
        self.progressbar.close()

    def __repr__(self):
        return self.__class__.__name__



from collections import deque
from typing import Iterable, Union

from .resource import Resources
from .task import Task
from .utils import get_logger


class Stage(object):

    '''
    One stage in a pipeline. A stage can have multiple peer worker processes.
    Worker number is desided based on available resources.
    If no resource is provided for allocation, defaults to max_worker or 1.
    '''

    def __init__(self, resources: Union[None, Resources] = None,
                 max_worker: Union[None, int] = None,
                 result_queue_size: int = 32):
        self.resources = resources
        if resources is not None:
            self.resource_allocation = resources.allocate(
                self.get_resource_unit())
            self.worker_num = len(self.resource_allocation)
            if max_worker is not None:
                self.worker_num = min(self.worker_num, max_worker)
                self.resource_allocation = self.resource_allocation[
                    :self.worker_num]
        else:
            if max_worker is None:
                max_worker = 1
            self.resource_allocation = [{} for _ in range(max_worker)]
            self.worker_num = max_worker
        self.result_queue_size = result_queue_size
        self.logger = None

    def get_resource_unit(self):
        '''
        Resource requirement for each worker.
        '''
        return {'cpu': 0.1}

    def reset(self):
        '''
        Reset a worker process. Automatically executed during initialization.
        '''
        self.logger.debug('Reset.')

    def process(self, task: Task) -> Union[Task, Iterable[Task]]:
        '''
        Process function for each worker process. 
        Returns one or a series of downstream tasks.
        '''
        raise NotImplementedError

    def init(self, worker_id: int = 0):
        self.worker_id = worker_id
        self.current_resource = self.resource_allocation[worker_id]
        name = '%s[%d]%s' % (self.__class__.__name__, worker_id,
                             '@' + str(self.current_resource)
                             if self.current_resource is not None else '')
        self.logger = get_logger(name)
        self.logger.debug('Init.')
        self.reset()

    def run(self, task):
        return self.process(task)

    def __repr__(self):
        return '%s(x%d)%s' % (
            self.__class__.__name__, self.worker_num,
            '@' + str(self.resources) if self.resources is not None else '')


class ReorderStage(Stage):

    '''
    A special stage with only one worker. 
    Tasks are processed according to a pre-defined order.
    '''

    def __init__(self, resources: Union[None, Resources] = None,
                 result_queue_size: int = 32):
        super(ReorderStage, self).__init__(resources, 1, result_queue_size)

    def get_sequence_id(self, task: Task) -> int:
        '''
        Return the order of each task. Start from 0.
        '''
        raise NotImplementedError

    def reset(self):
        super(ReorderStage, self).reset()
        self.reorder_buffer = deque()
        self.next_id = 0

    def push(self, task: Task):
        offset = self.get_sequence_id(task) - self.next_id
        if offset >= len(self.reorder_buffer):
            self.reorder_buffer.extend([None] * (
                offset - len(self.reorder_buffer) + 1))
        self.reorder_buffer[offset] = task

    def iter_pop(self):
        while len(self.reorder_buffer) > 0:
            if self.reorder_buffer[0] is not None:
                task = self.reorder_buffer.popleft()
                yield task

    def run(self, task):
        self.push(task)
        for task in self.iter_pop():
            res = self.process(task)
            if isinstance(res, Task):
                yield res
            else:
                for r in res:
                    yield r



import os
from itertools import cycle
from typing import Any, Dict, List, Union

import GPUtil


class Resources(object):

    '''
    Resource management and allocation.
    '''

    def __init__(self, resources: Dict[str, List[Any]] = None, **scan_args):
        if resources is not None:
            self.resources = resources
        else:
            self.resources = self.scan(**scan_args)

    def scan(self, gpu_load: float = 0.1, gpu_memory: float = 0.1):
        resources = {'cpu': [*range(os.cpu_count())]}
        gpus = GPUtil.getAvailable(
            limit=100, maxLoad=gpu_load, maxMemory=gpu_memory)
        if len(gpus) > 0:
            resources['gpu'] = gpus
        return resources

    def split(self, num_split: int):
        results = [{} for _ in range(num_split)]
        for key, value in self.resources.items():
            split_size = len(value) // num_split
            for split_i in range(num_split):
                results[split_i][key] = value[
                    split_i * split_size: (split_i + 1) * split_size]
        results = [Resources(r) for r in results]
        return results

    def allocate(self, resource_unit: Dict[str, Union[int, float]]):
        num_split = min([int(len(self.resources[key]) / unit)
                         for key, unit in resource_unit.items()])
        results = [{} for _ in range(num_split)]
        for key, unit in resource_unit.items():
            value = self.resources[key]
            if isinstance(unit, int):
                for split_i in range(num_split):
                    results[split_i][key] = value[
                        split_i * unit: (split_i + 1) * unit]
            else:
                for split_i, v in zip(range(num_split), cycle(value)):
                    results[split_i][key] = [v]
        results = [Resources(r) for r in results]
        return results

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join(
            ['%s=%s' % (k, v) for k, v in self.resources.items()]))
